/*
 * Copyright (C) 2001 and 2002 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 *      ctmpl/vector.c
 *      $Id: vector.c,v 1.1.1.1 2004/01/03 15:41:41 hiddy Exp $
 *              written by Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 */

#include "ctmpl/memory.h"
#include "ctmpl/memory.c"

static CTMPL_INLINE void 
PKG_NAME_(vector_set_ctl)(PKG_NAME_(vector) lp);

/* confirm size of vector */
static CTMPL_INLINE void 
PKG_NAME_(vec_confirm)(PKG_NAME_(vector) v, int i, int size);
static CTMPL_INLINE void 
PKG_NAME_(vec_insertAt)(PKG_NAME_(vector) v, 
			 unsigned int idx, 
			 PKG_NAME_(vecdata_t) dp);
static CTMPL_INLINE void 
PKG_NAME_(vec_insertsAt)(PKG_NAME_(vector) v, 
			 unsigned int idx, 
			 PKG_NAME_(vecdata_t) *dp, 
			 unsigned int size);
static CTMPL_INLINE PKG_NAME_(vecdata_t) 
PKG_NAME_(vec_removeAt)(PKG_NAME_(vector) v, 
		 	unsigned int idx);

/*   initialized vector size & expand size */
CTMPL_STATIC CTMPL_INLINE PKG_NAME_(vector) 
PKG_NAME_(vector_new)(int size, int esize)
{
	return PKG_NAME_(vector_new_by)(size, esize, NULL, NULL);
}
CTMPL_STATIC CTMPL_INLINE PKG_NAME_(vector) 
PKG_NAME_(vector_new_by)(int size, 
		      int esize, 
		      m_alloc_func_t af, 
		      m_free_func_t ff)
{
	int allocsize = 0;
	PKG_NAME_(vector) ret;
	
	ctmpl_assert(size > 0);
	ctmpl_assert(esize > 0);

	/* default alloc/free function */
	if(!af)	af = PKG_NAME_(m_alloc);
	if(!ff)	ff = PKG_NAME_(m_free);
		
	ret = CTMPL_MALLOC(af, PKG_NAME_STRING "vector", 
				sizeof(struct PKG_NAME_(vector)));

	allocsize = sizeof(PKG_NAME_(vecdata_t)) * size;
	ret->array = CTMPL_MALLOC(af, PKG_NAME_STRING "vector->array",  	
				  allocsize);

	ret->size = 0; /* current size */
	ret->capacity = size;	/* capacity of array */
	ret->expandsize = esize; /* expand size */
	ret->alloc = af;
	ret->free = ff;
	PKG_NAME_(vector_set_ctl)(ret);
	
	return ret;
}

static CTMPL_INLINE void 
PKG_NAME_(vec_clear_by)(PKG_NAME_(vector) v, 
			PKG_NAME_(vecdata_access_func_t) f)
{
	int i = 0;
	ctmpl_assert(v != NULL);

	if(f != NULL)
		for(i = 0; i < v_size(v); i++)
			(*f)(v_getAt(v, i));
	memset(v->array, 0, 
		sizeof(PKG_NAME_(vecdata_t)) * v->capacity);
	v_size(v) = 0;
}

static CTMPL_INLINE void 
PKG_NAME_(vec_delete_by)(PKG_NAME_(vector) v, 
			 PKG_NAME_(vecdata_access_func_t) f)
{
	PKG_NAME_(vec_clear_by)(v, f);
	CTMPL_FREE(v->free, PKG_NAME_STRING "vector->array", v->array);
	CTMPL_FREE(v->free, PKG_NAME_STRING "vector", v);
}

static CTMPL_INLINE PKG_NAME_(vector) 
PKG_NAME_(vec_dup_by)(PKG_NAME_(vector) v,
			 PKG_NAME_(vecdata_duplicate_func_t) f)
{
	int i;
	PKG_NAME_(vector) ret = 
		PKG_NAME_(vector_new_by)(v->capacity, v->expandsize,
				      v->alloc, v->free);
	ret->size = v->size;
	for(i = 0; i < v->size; i++)
		ret->array[i] = (f? f(v->array[i]): v->array[i]);
	return ret;
}

static CTMPL_INLINE void 
PKG_NAME_(vec_foreach)(PKG_NAME_(vector) v, 
		       PKG_NAME_(vecdata_foreach_func_t) f,
		       void *arg)
{
	int i = 0;
	ctmpl_assert(v != NULL);

	if(f != NULL)
		for(i = 0; i < v_size(v); i++)
			(*f)(v_getAt(v, i), arg);
}


static CTMPL_INLINE iterator_cnt_t
PKG_NAME_(vec_count)(PKG_NAME_(vector) v) { return v_size(v); }


static CTMPL_INLINE PKG_NAME_(vector_iterator) 
PKG_NAME_(vec_itr_first)(PKG_NAME_(vector) lp)
{
	PKG_NAME_(vector_iterator) ret;
	iterator_settype(ret, CTMPL_ITER_FORWARD);
	iterator_subtype(ret) = 0;
	iterator_aggregate(ret) = lp;
	iterator_item(ret) = (v_size(lp) > 0 ? 0: -1);
	return ret;
}

static CTMPL_INLINE PKG_NAME_(vector_iterator) 
PKG_NAME_(vec_itr_last)(PKG_NAME_(vector) lp)
{
	PKG_NAME_(vector_iterator) ret;
	iterator_settype(ret, CTMPL_ITER_FORWARD);
	iterator_subtype(ret) = 0;
	iterator_aggregate(ret) = lp;
	iterator_item(ret) = (v_size(lp) > 0 ? v_size(lp)-1: -1); 
	return ret;
}

static CTMPL_INLINE PKG_NAME_(vector_iterator) 
PKG_NAME_(vec_itr_at)(PKG_NAME_(vector) lp, iterator_pos_t idx)
{
	int prev = 0;
	PKG_NAME_(vector_iterator) ret;
	iterator_settype(ret, CTMPL_ITER_FORWARD);
	iterator_subtype(ret) = 0;
	iterator_aggregate(ret) = lp;

	if(idx < 0){
		++prev;
		idx = -idx - 1;
	}
	if(idx >= v_size(lp))
		iterator_item(ret) = -1;
	else {
		idx = (prev? v_size(lp) - idx: idx); 
		iterator_item(ret) = idx;
	}
	return ret;
}

static CTMPL_INLINE PKG_NAME_(vector_iterator) 
PKG_NAME_(vec_itr_next)(PKG_NAME_(vector_iterator) itr)
{
	PKG_NAME_(vector_iterator) ret;
	iterator_settype(ret, iterator_type(itr));
	iterator_subtype(ret) = 0;
	iterator_aggregate(ret) = iterator_aggregate(itr);
	iterator_item(ret) = -1;

	if(!iterator_is_end(itr)){
		PKG_NAME_(vector) v = iterator_aggregate(itr);
		if(iterator_item(itr) < (v_size(v) - 1))
			iterator_item(ret) = iterator_item(itr) + 1;
	}
	return ret;
}

static CTMPL_INLINE PKG_NAME_(vector_iterator) 
PKG_NAME_(vec_itr_prev)(PKG_NAME_(vector_iterator) itr)
{
	PKG_NAME_(vector_iterator) ret;
	iterator_settype(ret, iterator_type(itr));
	iterator_subtype(ret) = iterator_subtype(itr); 
	iterator_aggregate(ret) = iterator_aggregate(itr);
	iterator_item(ret) = -1;

	if(!iterator_is_end(itr)){
		if(iterator_item(itr) > 0)
			iterator_item(ret) = iterator_item(itr) - 1;
	}
	return ret;
}

static CTMPL_INLINE PKG_NAME_(vector_iterator) 
PKG_NAME_(vec_itr_erase_gonext)(PKG_NAME_(vector_iterator) itr)
{
	PKG_NAME_(vector_iterator) ret;
	PKG_NAME_(vector) v = iterator_aggregate(itr);
	iterator_settype(ret, iterator_type(itr));
	iterator_subtype(ret) = iterator_subtype(itr); 
	iterator_aggregate(ret) = iterator_aggregate(itr);
	iterator_item(ret) = -1;

	/* no item to erase */
	if(iterator_is_end(itr))
		return ret;

	/* this element is last elm of vector */
	if(iterator_item(itr) < (v_size(v)-1))
		iterator_item(ret) = iterator_item(itr) + 1;

	(void)PKG_NAME_(vec_removeAt)(iterator_aggregate(ret), 
				      iterator_item(itr));
	return ret;
}

static CTMPL_INLINE PKG_NAME_(vector_iterator) 
PKG_NAME_(vec_itr_erase_goprev)(PKG_NAME_(vector_iterator) itr)
{
	PKG_NAME_(vector_iterator) ret;
	iterator_settype(ret, iterator_type(itr));
	iterator_subtype(ret) = iterator_subtype(itr); 
	iterator_aggregate(ret) = iterator_aggregate(itr);
	iterator_item(ret) = -1;

	/* no item to erase */
	if(iterator_is_end(itr))
		return ret;

	/* this element is not first elm of vector */
	if(iterator_item(itr) > 0)
		iterator_item(ret) = iterator_item(itr) - 1;

	(void)PKG_NAME_(vec_removeAt)(iterator_aggregate(ret), 
				      iterator_item(itr));
	return ret;
}

static int 
PKG_NAME_(vec_itr_is_equal)(PKG_NAME_(vector_iterator) x,
                            PKG_NAME_(vector_iterator) y)
{ return CTMPL_ITR_IS_EQUAL(x, y); }


static CTMPL_INLINE PKG_NAME_(vecdata_t)
PKG_NAME_(vec_itr_retrieve)(PKG_NAME_(vector_iterator) itr)
{
	PKG_NAME_(vector) v = iterator_aggregate(itr);
	return v->array[iterator_item(itr)];
}

static CTMPL_INLINE PKG_NAME_(vector_iterator) 
PKG_NAME_(vec_find_between_by)(PKG_NAME_(vector_iterator) beg,
				  PKG_NAME_(vector_iterator) end,
				  PKG_NAME_(vecdata_t) dat,
				  PKG_NAME_(vecdata_compare_func_t) cmp)
{
	PKG_NAME_(vector_iterator) itr, ret;
	iterator_settype(ret, iterator_type(beg));
	iterator_subtype(ret) = iterator_subtype(beg); 
	iterator_aggregate(ret) = iterator_aggregate(beg);
	iterator_item(ret) = -1;
	
	itr = beg;

	while(!iterator_is_end(itr) &&
	      iterator_item(itr) != iterator_item(end)){
		PKG_NAME_(vecdata_t) dat2 = iterator_data(itr);
		/* is same data? */
		if((cmp? !cmp(dat, dat2): dat == dat2))
			return itr;

		itr = iterator_next(itr);
	}
	/* not found */
	return ret;
}

static CTMPL_INLINE PKG_NAME_(vector_iterator) 
PKG_NAME_(vec_find)(PKG_NAME_(vector) lp, 
		    PKG_NAME_(vecdata_t) dat)
{
	return PKG_NAME_(vec_find_between_by) (iterator_first(lp), 
				iterator_end(lp), dat, NULL);
}

#define IS_BETWEEN(l, t) \
	 ((l->array <= iterator_item(t)) && \
	  (iterator_item(t) < &l->array[l->size]))

static CTMPL_INLINE void 
PKG_NAME_(vec_insert_after)(PKG_NAME_(vector_iterator) itr, 
			    PKG_NAME_(vecdata_t) v)
{
	PKG_NAME_(vector) l = iterator_aggregate(itr);
	int 	isEnd = iterator_is_end(itr);
	unsigned int diff = 0;
	ctmpl_assert(l != NULL);

	if(!isEnd)
		diff = iterator_item(itr) + 1;
	PKG_NAME_(vec_insertAt)(l, diff, v); /* after */
}

static CTMPL_INLINE void 
PKG_NAME_(vec_insert_before)(PKG_NAME_(vector_iterator) itr, 
			      PKG_NAME_(vecdata_t) v)
{
	PKG_NAME_(vector) l = iterator_aggregate(itr);
	int 	isEnd = iterator_is_end(itr);
	unsigned int diff = 0;
	ctmpl_assert(l != NULL);

	if(!isEnd)
		diff = iterator_item(itr);
	PKG_NAME_(vec_insertAt)(l, diff, v); /* before */
}

static CTMPL_INLINE void
PKG_NAME_(vec_swap)(PKG_NAME_(vector_iterator) itr1, 
		    PKG_NAME_(vector_iterator) itr2)
{
	PKG_NAME_(vector) l = iterator_aggregate(itr1);
	PKG_NAME_(vecdata_t) v;
	ctmpl_assert(!iterator_is_end(itr1) && !iterator_is_end(itr2));

	v = l->array[iterator_item(itr1)];
	l->array[iterator_item(itr1)] = l->array[iterator_item(itr2)]; 
	l->array[iterator_item(itr2)] = v;
}

static CTMPL_INLINE PKG_NAME_(vecdata_t)
PKG_NAME_(vec_assign)(PKG_NAME_(vector_iterator) itr, 
		      PKG_NAME_(vecdata_t) dat)
{
	PKG_NAME_(vector) l = iterator_aggregate(itr);
	ctmpl_assert(!iterator_is_end(itr));

	l->array[iterator_item(itr)] = dat;
	return dat;
}

static CTMPL_INLINE void 
PKG_NAME_(vec_insertAt)(PKG_NAME_(vector) v, 
			unsigned int i, 
		 	PKG_NAME_(vecdata_t) d)
{
	PKG_NAME_(vec_insertsAt)(v, i, &d, 1);
}

static CTMPL_INLINE void 
PKG_NAME_(vec_insertsAt)(PKG_NAME_(vector) v, 
			 unsigned int idx, 
			 PKG_NAME_(vecdata_t) *dp, 
			 unsigned int size)
{
	int i;
	ctmpl_assert(v != NULL);
	ctmpl_assert(idx <= v->size); 
	ctmpl_assert(dp != NULL);

	if(size == 0)   return;
	
	PKG_NAME_(vec_confirm)(v, idx, size);
	for(i = 0; i < size; i++)
		v->array[idx + i] = dp[i];
	v->size += size;
}

static CTMPL_INLINE PKG_NAME_(vecdata_t) 
PKG_NAME_(vec_removeAt)(PKG_NAME_(vector) v, 
		 	unsigned int idx)
{
	register int i;
	PKG_NAME_(vecdata_t) ret;

	ret = v_getAt(v, idx);
	for(i = idx; i < v->size - 1; i++)
		v_getAt(v, i) = v_getAt(v, i+1);
		
	v->size--;
	return ret;
}	

static CTMPL_INLINE void 
PKG_NAME_(vec_confirm)(PKG_NAME_(vector) v, 
		       int idx, 
		       int size)
{
	int i = 0;
	int newsize = 0;
	PKG_NAME_(vecdata_t) *newarray; 

	/* calling function has to check v & index i */

	/* over capacity, calculate size to expand */	
	if(v->size + size > v->capacity){
		int allocsize;
		for(i = 1; ; i++){
			newsize = v->capacity + (i * v->expandsize);
			if(v->size + size <= newsize)
				break;
		}

		allocsize = sizeof(PKG_NAME_(vecdata_t)) * newsize;
		newarray = CTMPL_MALLOC(v->alloc, 
					PKG_NAME_STRING "vector->array",  
					allocsize);
	}
	else
		newarray = v->array;	/* enough size */

	if(newarray != v->array && idx > 0)
		for(i = 0; i < idx; i++)
			newarray[i] = v->array[i];

	/* copy from end of array */
	if(idx < v->size) 
		for(i = v->size - 1; i >= idx; i--)
			newarray[size + i] = v->array[i];

	/* free old array */
	if(newarray != v->array){
		CTMPL_FREE(v->free, PKG_NAME_STRING "vector->array",  v->array);
		v->array = newarray;
		v->capacity = newsize;
	}
}

static PKG_NAME_(vector_ctl) PKG_NAME_(vector_ctl_sw) = 
{
	/* end iterator */
	{ -1, -1,  NULL, -1 },
	PKG_NAME_(vec_delete_by),
	PKG_NAME_(vec_clear_by),
	PKG_NAME_(vec_dup_by),
	PKG_NAME_(vec_foreach),
	PKG_NAME_(vec_count),
	PKG_NAME_(vec_itr_first),
	PKG_NAME_(vec_itr_last),
	PKG_NAME_(vec_itr_at),
	PKG_NAME_(vec_itr_next),
	PKG_NAME_(vec_itr_prev),
	PKG_NAME_(vec_itr_erase_gonext),
	PKG_NAME_(vec_itr_erase_goprev),
	PKG_NAME_(vec_itr_is_equal),
	PKG_NAME_(vec_itr_retrieve),
	PKG_NAME_(vec_find),
	PKG_NAME_(vec_find_between_by),
	PKG_NAME_(vec_insert_before),
	PKG_NAME_(vec_insert_after),
	PKG_NAME_(vec_swap),
	PKG_NAME_(vec_assign)
};

/* set iterator controller */
static CTMPL_INLINE void 
PKG_NAME_(vector_set_ctl)(PKG_NAME_(vector) lp)
{
	lp->ctl = &PKG_NAME_(vector_ctl_sw);
}

