/*
 *			memory.c
 * $Id: memory.c,v 1.6 2004/01/22 21:58:14 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static void *mp_malloc(const char *, size_t, const char *, int);
static void mp_free(const char *, void *, const char *, int);
static void mp_dump(void);

/* set default allocator */
void *(*xpdic_malloc)(const char *, size_t, const char *, int) 
		=  mp_malloc;
void (*xpdic_free)(const char *, void *, const char *, int) 
		= mp_free;
void (*memdump)(void) = mp_dump;

#ifdef MEM_DEBUG

#ifndef MEM_DEBUG_NPIECE
#define MEM_DEBUG_NPIECE 	1024 * 1024
#endif /* MEM_DEBUG_NPIECE */

struct mem_piece {
	struct mem_piece *mp_next;
	const char 	 *mp_purpose;
	const char 	 *mp_filename;
	const void	 *mp_addr;
	int	 	 mp_size;
};
static struct mem_piece mpieces[MEM_DEBUG_NPIECE], 
			*freeptr, *usedptr;
static int inited = 0, npieces = 0;
static unsigned int peekbytes = 0, curbytes = 0;
static unsigned int peekps = 0;

static void 
mem_init(void)
{
	int i = 0;
	usedptr = NULL;
	freeptr = &mpieces[0];
	for(; i < MEM_DEBUG_NPIECE - 1; ++i)
		mpieces[i].mp_next = &mpieces[i+1];
	++inited; 
}

static void
add_mp(const char *purpose, 
       const char *filename, 
       const void *maddr, 
       int size)
{
	struct mem_piece *mp;
	if(!inited)
		mem_init();

	curbytes += size;
	peekbytes = (curbytes > peekbytes)? curbytes: peekbytes;

	if(!freeptr){
		fprintf(stderr, "MEM_DEBUG: no more memory pieces!!\n");
		abort();
		return;
	}
	mp = freeptr;
	freeptr = mp->mp_next;

	mp->mp_purpose = purpose;
	mp->mp_filename = filename;
	mp->mp_addr = maddr;
	mp->mp_size = size;

	mp->mp_next = usedptr;
	usedptr = mp;
	++npieces;
	peekps = (npieces > peekps)? npieces: peekps;
}

static void
del_mp(const char *p, 
       const char *filename, 
       const void *m, 
       int line)
{
	struct mem_piece *mp;
	if(!inited){
		fprintf(stderr, "MEM_DEBUG: not initialized in del_mp().\n");
		exit(1);
	}

	if(usedptr){
		struct mem_piece *prev;
		mp = prev = usedptr;
		for(; mp ; mp = mp->mp_next){
			if(mp->mp_addr == m)
				break;
			prev = mp;
		}
		if(mp){
			curbytes -= mp->mp_size;
			if(mp == usedptr)
				usedptr = mp->mp_next;
			else
				prev->mp_next = mp->mp_next;
			mp->mp_next = freeptr;
			freeptr = mp;
			mp->mp_purpose = mp->mp_filename = NULL;
			mp->mp_addr = NULL;
			mp->mp_size = 0;
			--npieces;
			return;
		}
	}
	fprintf(stderr, "MEM_DEBUG: %s@0x%x(%s:%d): no such piece.\n",
		p, (unsigned int)m, filename, line);
	abort();

}

static void
mp_dump(void)
{
	int 			i;
	struct mem_piece 	*mp = usedptr;
	fprintf(stderr, "dump_mp(): %d/%d pieces, %d/%d bytes\n", 
		npieces, peekps, curbytes, peekbytes);

	for(i = 0; mp ; mp = mp->mp_next, ++i){
		fprintf(stderr, "%d. %s@0x%x(%s): %d bytes\n", i, 
			mp->mp_purpose, (unsigned int)mp->mp_addr,
			mp->mp_filename, mp->mp_size);
	}
}
#else 
static void mp_dump(void) { }
#endif /* MEM_DEBUG */


static void *
mp_malloc(const char *purpose, size_t size, 
	  const char *filename, int line)
{
	void *ret = (void *)malloc(size);     
	if(ret == NULL){
		fprintf(stderr, 
		"mp_malloc: failed %d bytes for \'%s\' at %s:%d.\n", 
				size, purpose, filename, line);
		exit(1);
	}
#ifdef MEM_DEBUG
	add_mp(purpose, filename, ret, size);
#endif /* MEM_DEBUG */
	memset(ret, 0, size);
	return ret;
}

static void 
mp_free(const char *purpose, void *maddr,
	const char *filename, int line)
{
#ifdef MEM_DEBUG
	del_mp(purpose, filename, maddr, line);
#endif /* MEM_DEBUG */
	free(maddr);
}

