/*
 *			property.c
 * $Id: property.c,v 1.4 2004/01/19 22:16:19 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "memory.h"
#include "property.h"
#include "prptdb.c"

/************************************/
/* property database */
static prptdb _prptdb;
/************************************/

static int (*prptdb_cmp)(const char *c1, const char *c2) = strcasecmp;
static unsigned int prptdb_hash(const char *c1){
	unsigned int ret = 0;
	register int i = 0;

	for(i = 0; i < 4; ++i){
		if(c1[i] == '\0')
			break;
		ret += tolower(c1[i]);
	}
	return ret;
}

static void 
prptdb_data_free(const char *cp, struct property *p)
{
	property_free(p);
}

void 
property_init(void)
{
	if(!_prptdb)
		_prptdb = prptdb_new(13, prptdb_hash, prptdb_cmp,
				    xpdic_malloc, xpdic_free);
}

void 
property_fin(void)
{
	aggregator_delete_by(_prptdb, prptdb_data_free);	
}

void 
property_dump(void)
{
	register int i = 0;
	int ival = 0;
	prptdb_iterator itr;

	itr = iterator_first(_prptdb);

	for(; !iterator_is_end(itr); itr = iterator_next(itr), ++i){
		printf("%2d: %s = ", i, iterator_data(itr)->second->prty_name);

		switch(iterator_data(itr)->second->prty_type){
		case PRTY_TYPE_INTEGER:
			ival = prty_integer_val(iterator_data(itr)->second);
			if(ival >= -1 && ival < 1000)
				printf("%d\n", prty_integer_val(iterator_data(itr)->second));
			else 
				printf("0x%x\n", prty_integer_val(iterator_data(itr)->second));
			break;
		case PRTY_TYPE_STRING:
			printf("%s\n", prty_string_val(iterator_data(itr)->second));
		}
	}
}

struct property *
property_put(struct property *p)
{
	struct property *ret = NULL;
	prptdb_iterator itr = iterator_find(_prptdb, p->prty_name);

	if(!iterator_is_end(itr)){
		ret = iterator_data(itr)->second;
		iterator_erase(itr);
	}
	iterator_insert_pair(_prptdb, p->prty_name, p);
	return ret;
}

struct property *
property_get(const char *name)
{
	prptdb_iterator itr = iterator_find(_prptdb, name);
	if(iterator_is_end(itr))
		return NULL;
	return iterator_data(itr)->second; /* value */
}

static void
prpt_set_data(struct property *prp, int type, const char *name, void (*func)(struct property *))
{
	prp->prty_type = type;
	prp->prty_name = (const char *)xp_strdup(name);
	prp->prty_free = func;
}


static void 
prpt_free(struct property *p){ 
	xp_strfree(p->prty_name); 
}
static void 
prpt_integer_free(struct property *p)
{ 
	prpt_free(p);
	FREE("property/integer", p); 
}
static void 
prpt_string_free(struct property *p)
{ 
	prpt_free(p);
	xp_strfree(prty_string_val(p));
	FREE("property/string", p);
}


struct property *
property_integer_alloc(const char *name, int val)
{
	struct property *ret;
	if(!name)
		return NULL;
	ret = MALLOC("property/integer", sizeof(struct property_integer));
	prpt_set_data(ret, PRTY_TYPE_INTEGER, name, prpt_integer_free);
	prty_integer_val(ret) = val;
	return ret;
}
struct property *
property_string_alloc(const char *name, const char *val)
{
	struct property *ret;
	if(!name || !val)
		return NULL;
	ret = MALLOC("property/string", sizeof(struct property_string));
	prpt_set_data(ret, PRTY_TYPE_STRING, name, prpt_string_free);
	prty_string_val(ret) = xp_strdup(val);
	return ret;
}


#ifdef PROP_TEST

int
main(void)
{
	char buf[1024];
	char *p;
	char *namep, *valp;
	int  val;
	struct property *prop;

	property_init();
	while((p = fgets(buf, 1024, stdin)) != NULL){
		buf[strlen(buf)-1] = '\0';
		switch(*p){
		case 'l':
			property_dump();	
			break;
		case 'p':
			namep = p + 2;
			valp = index(namep, ':');
			*valp = '\0';
			valp++;
			printf("put: %s -> %s\n", namep, valp);
			if(*valp >= '0' && *valp <= '9')
				property_put(property_integer_alloc(namep, atoi(valp)));
			else 
				property_put(property_string_alloc(namep, valp));
			break;
		case 'g':
			namep = p + 2;
			printf("get(%s): ", namep);
			prop = property_get(namep);
			if(!prop){
				printf("no such property\n");
				break;
			}
			printf("%s -> ", prop->prty_name);
			if(prop->prty_type == PRTY_TYPE_INTEGER)
				printf("%d\n", prty_integer_val(prop));
			else
				printf("%d\n", prty_string_val(prop));
		}
	}
	property_fin();
	memdump();
	return 0;
}
#endif /* PROP_TEST */
