/*
 *			xactor.c
 * $Id: xactor.c,v 1.4 2004/01/21 20:08:13 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#define _XACTOR_INTERNAL
#include <stdio.h>
#include "memory.h"
#include "bitmap.h"
#include "xwindow.h"
#include "xobject.h"
#include "xsignal.h"
#include "xactor.h"

struct xaction *
xaction_new(xobject *x, xaction_t a, void *v)
{
	struct xaction *ret = MALLOC("xaction", 
					sizeof(struct xaction));
	if(x)
		xobj_incref(x);
	ret->xac_object = x;
	ret->xac_action = a;
	ret->xac_ptr = v;
	return ret;
}

void 
xaction_delete(struct xaction *r)
{
	if(r->xac_object)
		xobj_free(r->xac_object);
	FREE("xaction", r);
}


void
xact_send_signal(struct xactor *a, xpoint pos, xsignal sig, struct xsigarg *sigarg, 
		 struct xobject *root, xwindow *xwp)
{
	struct xaction *xact = NULL;

	/* check if the signal is acceptable for the actor */
	if(!xact_acceptable(a, sig)){
		fprintf(stderr, "xactor_signal():  can't send signal(%d) to unacceptable actor<0x%x>.\n", sig, (int)a); 
		return;
	}

	if(!xobj_is_active(a)){
		fprintf(stderr, "xactor_signal():  can't send signal(%d) to inactive actor<0x%x>.\n", sig, (int)a); 
		return;
	}

	/* get xaction for the signal */
	xact = (a->xa_getaction)(a, sig);

	/* send the signal to the actor */
	(a->xa_receive)(a, pos, sig, sigarg, root, xwp);

	if(xact && xact->xac_object){
		xobject *save = a->xa_object;
		a->xa_object = xact->xac_object;
		xobj_redraw_object(root, a, pos.x, pos.y, xwp, 0);
		a->xa_object = save;
	}

	/* execute the registered action for this signal */
	if(xact && xact->xac_action)
		(xact->xac_action)(a, pos, sig, sigarg, xact->xac_ptr,
				   root, xwp);
}


void
xactor_draw(struct xobject *o, int x, int y, xwindow *xwp, 
	    int rx, int ry, unsigned int rw, unsigned int rh, int f)
{
	struct xactor *xrp = (struct xactor *)o;
	xobj_draw_rect(xrp->xa_object, x, y, xwp, rx, ry, rw, rh, 
			xobj_is_active(o)? f : f | XOBJDRAW_INACTIVE);
}

void 
xactor_relocate(struct xcomposer *o, int x, int y)
{
	struct xactor *xrp = (struct xactor *)o;
	if(xobj_is_composer(xrp->xa_object))
		xcmp_relocate(xrp->xa_object, x, y);
}

struct xobject *
xactor_find(struct xcomposer *o, int x, int y, int f, xpoint *p)
{ 
	struct xactor *r = (struct xactor *)o;
	struct xobject *ret = NULL;
	struct xpoint tmp = { 0, 0 };

	/* I'm not an active actor */
	if(!xobj_is_active(r))
		return NULL;

	/* requiring first matched object */
	if(xcmpfind_is_firstmatch(f)
	&& (xcmpfind_type(f) == XOBJTYPE_ACTOR || xcmpfind_typeany(f))
	&& r->xa_acceptable(r, xcmpfind_subtype(f)))
		return (struct xobject *)r;

	/* find lower objects */
	if(xobj_is_composer(r->xa_object))
		ret = xcmp_find(r->xa_object, x, y, f, &tmp);

	/* object found */
	if(ret){
		p->x += tmp.x;
		p->y += tmp.y;
		return ret;
	}

	if((xcmpfind_type(f) == XOBJTYPE_ACTOR || xcmpfind_typeany(f))
	&& r->xa_acceptable(r, xcmpfind_subtype(f)))
		return (struct xobject *)r;
	
	return NULL;
}

void 
xactor_init(struct xactor *r, struct xobject *o, 
	    x_draw_t draw, x_free_t free, x_dump_t dump, 
	     void (*compose)(struct xcomposer *o),
	     struct xobject *(*find)(struct xcomposer *o, int x, int y, int sig, xpoint *),
	     void (*relocate)(struct xcomposer *o, int x, int y))
{
	struct xobject *xp = xobjcast(r);
        xp->x_type = XOBJTYPE_ACTOR;
        xp->x_flags = XOBJFLAG_ENABLE | XOBJFLAG_ACTIVE;
	if(xcmp_wvariable(o))
        	xp->x_flags |= XOBJFLAG_WVARIABLE;
	if(xcmp_hvariable(o))
        	xp->x_flags |= XOBJFLAG_HVARIABLE;
        xp->x_width = xobj_width(o);
        xp->x_height = xobj_height(o);
        xp->x_draw = draw;
        xp->x_free = free;
        xp->x_dump = dump;
        xp->x_refcnt = 0;
	xcmpcast(r)->xc_compose = compose;
	xcmpcast(r)->xc_find = find;
	xcmpcast(r)->xc_relocate = relocate;
	r->xa_object = o;
	xobj_incref(o);
}

