/*
 *			xkeyaction.c
 * $Id: xkeyaction.c,v 1.3 2004/01/19 03:33:21 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#define _XACTOR_INTERNAL

#include <stdio.h>
#include "memory.h"
#include "bitmap.h"
#include "xwindow.h"
#include "xobject.h"
#include "xsignal.h"
#include "xactor.h"
#include "resource.h"

/*
 *    X Clickable Reactor Object 
 *	Treated signal: 
 *		- XSIG_ENTER
 *		- XSIG_LEAVE
 *		- XSIG_KEYPRESS
 *		- XSIG_KEYRELEASE
 */
struct xkeyaction {
	struct xactor 	xcl_base;
	struct xaction	*xcl_press;
	struct xaction	*xcl_release;
};


static void
xrfree(struct xobject *o)
{
	struct xkeyaction *c = (struct xkeyaction *)o;
	if(c->xcl_press)
		xaction_delete(c->xcl_press);
	if(c->xcl_release)
		xaction_delete(c->xcl_release);
	xobj_free(((struct xactor *)c)->xa_object);
	FREE("xkeyaction", c);
}

#define actdump(x, id, f, i, idt)  do {					\
	if(!x)								\
		fprintf(fp, "%s %s: not prepaired.\n", idt, id); 	\
	else if(!(x)->xac_object)					\
		fprintf(fp, "%s %s: no object to be drawn.\n", idt, id); \
	else {								\
		fprintf(fp, "%s %s:\n", idt, id); 			\
		xobj_dump((x)->xac_object, fp, i + 1);		\
	}								\
} while(0)

static void
xrdump(struct xobject *o, FILE *fp, int off)
{
	struct xkeyaction *c = (struct xkeyaction *)o;
	struct xactor *r = (struct xactor *)o;
	char idt[1024], buf[1024], info[1024], *p = idt;

	xod_indent(idt, off);
	xobj_info(o, info, 1024);
	
	snprintf(buf, 1024, "%s%s<Clickable>: %s{\n", 
		 idt, xobj_typestr(o), info);
	fprintf(fp, "%s", buf);

	while(*p)
		*p++ = ' ';

	actdump(c->xcl_press, "KeyPress", fp, off, idt); 
	actdump(c->xcl_release, "KeyRelease", fp, off, idt); 

	fprintf(fp, "%s Main:\n", idt); 
	xobj_dump(r->xa_object, fp, off + 1);
	fprintf(fp, "%s}\n", idt);
}

static void
xrcompose(struct xcomposer *o)
{
	struct xactor *r = (struct xactor *)o;
	struct xkeyaction *c = (struct xkeyaction *)o;

	if(xobj_is_composer(r->xa_object))
		xcmp_compose(r->xa_object);
	if(xcmp_wvariable(o))
		xobj_width(o) = xobj_width(r->xa_object);
	if(xcmp_hvariable(o))
		xobj_height(o) = xobj_height(r->xa_object);

	xact_compose(c->xcl_press);
	xact_compose(c->xcl_release);
}

static int
xracceptable(struct xactor *r, xsignal s)
{
	switch(xsignal_type(s)){
	case XSIG_KEYPRESS:
	case XSIG_KEYRELEASE:
		return 1;
	}
	return 0;

}

static struct xaction *
xrgetact(struct xactor *r, xsignal s)
{
	struct xkeyaction *c = (struct xkeyaction *)r;
	switch(xsignal_type(s)){
	case XSIG_KEYPRESS:
		return c->xcl_press;
	case XSIG_KEYRELEASE:
		return c->xcl_release;
	}
	return NULL;
}

static int
xrregister(struct xactor *r, xsignal s, xaction_t a, 
	   void *p, struct xobject *o)
{
	struct xkeyaction *c = (struct xkeyaction *)r;
	struct xaction **dst = NULL;

	switch(xsignal_type(s)){
	case XSIG_KEYPRESS:
		dst = &c->xcl_press;
		break;
	case XSIG_KEYRELEASE:
		dst = &c->xcl_release;
		break;
	}

	/* unacceptable signal */
	if(!dst)
		return 0;  /* fail */

	/* delete old action */
	if(*dst)
		xaction_delete(*dst);

	*dst = xaction_new(o, a, p);
	return 1;	/* success */
}


static void
xrreceive(struct xactor *act, xpoint pos, xsignal sig, struct xsigarg *sigarg, 
	  struct xobject *root, xwindow *w)
{
	if(xsignal_is_redirect(sig))
		return; /* do nothing */

	switch(xsignal_type(sig)){
	case XSIG_KEYPRESS:
		break;
	case XSIG_KEYRELEASE:
		break;
	}
}

struct xactor *
xkeyaction_new(struct xobject *o)
{
	struct xkeyaction *c = MALLOC("xkeyaction", sizeof(struct xkeyaction));
	struct xactor *ret = (struct xactor *)c;

	xactor_init(ret, o, xactor_draw, xrfree, xrdump, xrcompose, 
		      xactor_find, xactor_relocate);

	ret->xa_acceptable = xracceptable;
	ret->xa_getaction = xrgetact;
	ret->xa_register = xrregister;
	ret->xa_receive = xrreceive;
	c->xcl_press = NULL;
	c->xcl_release = NULL;
	return ret;
}
