/*
 *			xprimitive.c
 * $Id: xprimitive.c,v 1.11 2004/01/22 03:52:35 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include "memory.h"
#include "bitmap.h"
#include "xwindow.h"
#include "xobject.h"
#include "xprimitive.h"

#define xo_abs(x)	((x) < 0 ? -(x) : (x))
#define xo_max(x, y)	((x) < (y) ? (y) : (x))
#define xo_min(x, y)	((x) < (y) ? (x) : (y))


/* static draw functions */
static void xline_draw(struct xobject *xp, xwindow *xwp, int x, int y);
static void xrect_draw(struct xobject *xp, xwindow *xwp, int x, int y);
static void xarc_draw(struct xobject *xp, xwindow *xwp, int x, int y);
static void xbmp_draw(struct xobject *xp, xwindow *xwp, int x, int y);
static void xstr_draw(struct xobject *xp, xwindow *xwp, int x, int y);
static void xwstr_draw(struct xobject *xp, xwindow *xwp, int x, int y);
static void xeditstr_draw(struct xobject *xp, xwindow *xwp, int x, int y);
static void xprimitive_draw(struct xobject *xp, int x, int y, xwindow *xwp, 
			    int rx, int ry, unsigned int rw, unsigned int rh, 
			    int flag);

/*------------* X Primitive Function *-------------*/
static void 
xo_set(struct xobject *xp, int type, int flags, 
       unsigned int w, unsigned int h, 
       x_draw_t drawf, x_free_t freef, x_dump_t dumpf)
{
	xp->x_type = type;
	xp->x_flags = flags;
	xp->x_width = w;
	xp->x_height = h;
	xp->x_draw = xprimitive_draw;
	xp->x_free = freef;
	xp->x_dump = dumpf;
	xp->x_refcnt = 0;
}

static void 
xo_free(struct xobject *xp){ FREE(xobj_typestr(xp), xp); }
static void 
xo_dump(struct xobject *xp, FILE *fp, int off)
{  
	int  n = 0;
	char buf[1024], info[1024];
	n = xod_indent(buf, off);
	xobj_info(xp, info, 1024);
	snprintf(buf + n, 1024 - n, "%s: %s", xobj_typestr(xp), info);
	fprintf(fp, "%s", buf);
}

static void
check_region(struct xobject *xp, int x, int y, 
	     int rx, int ry, int rw, int rh) 
{
	int maxx = xobj_width(xp);	
	int maxy = xobj_height(xp);	

	if(maxx <= rx)
		fprintf(stderr, "xpmt_draw(%s): maxx(%d) < rx(%d)\n", 
			xobj_typestr(xp), maxx, rx);
	if(maxy <= ry)
		fprintf(stderr, "xpmt_draw(%s): maxy(%d) < ry(%d)\n", 
			xobj_typestr(xp), maxy, ry);
	if(rx + rw <= 0)
		fprintf(stderr, "xpmt_draw(%s): rx(%d) + rw(%d) < x(%d)\n", 
			xobj_typestr(xp), rx, rw, x);
	if(ry + rh <= 0)
		fprintf(stderr, "xpmt_draw(%s): ry(%d) + rh(%d) < y(%d)\n", 
			xobj_typestr(xp), ry, rh, y);

}

static void 
xprimitive_draw(struct xobject *xp, int x, int y, xwindow *xwp, 
		int rx, int ry, unsigned int rw, unsigned int rh, 
		int flag)
{
	long oldc;

	if(xobj_type(xp) == XOBJTYPE_EXPANDER ||
	   xobj_type(xp) == XOBJTYPE_SPACE)
		return;		/* no need to draw */

	/* 
        printf("xpmt_draw(%s): (x,y)(%d,%d)x(%d,%d), (rx,ry)=(%d,%d) (rw,rh)=(%d,%d)\n", 
				xobj_typestr(xp), x, y, xobj_width(xp), xobj_height(xp), rx, ry, rw, rh);
	*/

	/* check region size */
	check_region(xp, x, y, rx, ry, rw, rh);

	/* draw inactive object? */
	if(flag & XOBJDRAW_INACTIVE){
		/* save old foreground color */
		xattr_get_fgcolor(xpmt_attr(xp), &oldc);
		xattr_set_fgcolor(xpmt_attr(xp), XPMI_INACTIVE_COLOR);	
	}

	switch(xobj_type(xp)){
	case XOBJTYPE_LINE:
		xline_draw(xp, xwp, x, y);
		break;
	case XOBJTYPE_RECTANGLE:
		xrect_draw(xp, xwp, x, y);
		break;
	case XOBJTYPE_ARC:
		xarc_draw(xp, xwp, x, y);
		break;
	case XOBJTYPE_BITMAP:
		xbmp_draw(xp, xwp, x, y);
		break;
	case XOBJTYPE_STRING:
		xstr_draw(xp, xwp, x, y);
		break;
	case XOBJTYPE_WSTRING:
		xwstr_draw(xp, xwp, x, y);
		break;
	case XOBJTYPE_EDITSTR:
		xeditstr_draw(xp, xwp, x, y);
		break;
	}

	if(flag & XOBJDRAW_INACTIVE)
		xattr_set_fgcolor(xpmt_attr(xp), oldc);
}

/*-------------------* X Primitive Objects *---------------------*/
/*----- expander ----*/
static void xdump(struct xobject *xp, FILE *fp, int off)
{  
	xo_dump(xp, fp, off);
	fprintf(fp, "\n");
}

struct xobject *
xexpander_new(void)
{
	struct xobject *ret = MALLOC(xod_typestr(XOBJTYPE_EXPANDER), 
				     sizeof(struct xobject));
	xo_set(ret, XOBJTYPE_EXPANDER, XOBJFLAG_ENABLE, 0, 0,
		    NULL, xo_free, xdump);
	ret->x_flags |= XOBJFLAG_WVARIABLE;
	ret->x_flags |= XOBJFLAG_HVARIABLE;
	return ret;
}

/*----- space ----*/
struct xobject *
xspace_new(int width, int height)
{
	struct xobject *ret = MALLOC("xspace", sizeof(struct xobject));
	width = width < 0? 0: width;
	height = height < 0? 0: height;
	xo_set(ret, XOBJTYPE_SPACE, XOBJFLAG_ENABLE, width, height, 
		    NULL, xo_free, xdump);
	return ret;
}


/*----- line ----*/
static void xline_draw(struct xobject *xp, xwindow *xwp, int x, int y)
{
	struct xline *xlp = (struct xline *)xp;

	xw_draw_line(xwp, xpmt_attr(xp), xlp->x1 + x, xlp->y1 + y, 
		     xlp->x2 + x, xlp->y2 + y);
}

static void xline_dump(struct xobject *xp, FILE *fp, int oft) {  
	struct xline *xlp = (struct xline *)xp;
	xo_dump(xp, fp, oft);
	fprintf(fp, "{ x1=%d, y1=%d x2=%d y2=%d }\n", xlp->x1, xlp->y1, 
		xlp->x2, xlp->y2);
}

struct xobject *xline_new(int x1, int y1, int x2, int y2, xattr *at)
{
	int w, h;
	struct xline *ret = MALLOC(xod_typestr(XOBJTYPE_LINE), 
				   sizeof(struct xline));
	/* XXX: get line width */
	w = xo_max(x1, x2);
	h = xo_max(y1, y2);
	if(w == 0) w = 1;
	if(h == 0) h = 1;
	xo_set((struct xobject *)ret, XOBJTYPE_LINE, XOBJFLAG_ENABLE, 
			w, h, NULL, xo_free, xline_dump);
	xpmt_set_attr(ret, at);
	ret->x1 = x1;
	ret->y1 = y1;
	ret->x2 = x2;
	ret->y2 = y2;
	return (struct xobject *)ret;
}

/*----- rectangle ----*/
static void xrect_draw(struct xobject *xp, xwindow *xwp, int x, int y)
{
	struct xrect *xrp = (struct xrect *)xp;
	if(xrp->fill)
		xw_fill_rectangle(xwp, xpmt_attr(xp), x, y, xobj_width(xp), 
					xobj_height(xp));
	else
		xw_draw_rectangle(xwp, xpmt_attr(xp), x, y, xobj_width(xp), 
					xobj_height(xp));
}
static void xrect_dump(struct xobject *xp, FILE *fp, int oft) {  
	struct xrect *xlp = (struct xrect *)xp;
	xo_dump(xp, fp, oft);
	fprintf(fp, "{ fill=%d }\n", xlp->fill);
}

struct xobject *
xrect_new(int w, int h, int fill, xattr *at)
{
	struct xrect *ret = MALLOC(xod_typestr(XOBJTYPE_RECTANGLE), 
				   sizeof(struct xrect));
	xo_set((struct xobject *)ret, XOBJTYPE_RECTANGLE, 
		    XOBJFLAG_ENABLE, w, h, NULL, xo_free, xrect_dump);
	ret->fill = fill;
	xpmt_set_attr(ret, at);
	return (struct xobject *)ret;
}

/*----- arc ----*/
static void xarc_draw(struct xobject *xp, xwindow *xwp, int x, int y)
{
	struct xarc *xrp = (struct xarc *)xp;
	if(xrp->fill)
		xw_fill_arc(xwp, xpmt_attr(xp), x, y, xobj_width(xp), xobj_height(xp),
			    xrp->angle1, xrp->angle2);
	else
		xw_draw_arc(xwp, xpmt_attr(xp), x, y, xobj_width(xp), xobj_height(xp),
			    xrp->angle1, xrp->angle2);
}
static void xarc_dump(struct xobject *xp, FILE *fp, int oft) {  
	struct xarc *xlp = (struct xarc *)xp;
	xo_dump(xp, fp, oft);
	fprintf(fp, "{ angle1=%d angle2=%d fill=%d }\n", xlp->angle1, xlp->angle2, 
		xlp->fill);
}

struct xobject *
xarc_new(int w, int h, int a1, int a2, int fill, xattr *at)
{
	struct xarc *ret = MALLOC("xarc", sizeof(struct xarc));
	xo_set((struct xobject *)ret, XOBJTYPE_ARC, XOBJFLAG_ENABLE, w, h, 
		    NULL, xo_free, xarc_dump);
	ret->fill = fill;
	ret->angle1 = a1;
	ret->angle2 = a2;
	xpmt_set_attr(ret, at);
	return (struct xobject *)ret;
}

/*----- bitmap ----*/
static void xbmp_draw(struct xobject *xp, xwindow *xwp, int x, int y)
{
	struct xbitmap *xrp = (struct xbitmap *)xp;
	xw_draw_image(xwp, xpmt_attr(xp), x, y, xrp->image, 
			xobj_width(xp), xobj_height(xp));
}
static void xbmp_free(struct xobject *xp){ 
	struct xbitmap *xrp = (struct xbitmap *)xp;
	ximage_free(xrp->image);
	xrp->image = NULL;
	FREE("xbmp", xp); 
}

/* XXXX: should be changed not to recognise internal data structure */
struct _xatr { xrootwindow *rwp; };
struct xobject *xbitmap_new(struct bitmap *bmp, struct xattr *at)
{
	struct _xatr *atrp = (struct _xatr *)at; 
	struct xbitmap *ret = MALLOC(xod_typestr(XOBJTYPE_BITMAP), 
				  sizeof(struct xbitmap));
	xo_set((struct xobject *)ret, XOBJTYPE_BITMAP, XOBJFLAG_ENABLE, 
		bmp->bm_width, bmp->bm_height, NULL, xbmp_free, xdump);
	xpmt_set_attr(ret, at);
	ret->image = ximage_from_bitmap(atrp->rwp, bmp);
	ret->bmp = bmp;
	return (struct xobject *)ret;
}

/*----- string ----*/
static void xstr_draw(struct xobject *xp, xwindow *xwp, int x, int y)
{
	struct xstr *xrp = (struct xstr *)xp;
	if(xrp->bg)
		xw_draw_string_bg(xwp, xpmt_attr(xp), x, y, 
			xrp->str, xrp->nchars);
	else
		xw_draw_string(xwp, xpmt_attr(xp), x, y, 
			xrp->str, xrp->nchars);
}
static void xstr_dump(struct xobject *xp, FILE *fp, int oft) {  
	char *cp;
	struct xstr *xlp = (struct xstr *)xp;
	xo_dump(xp, fp, oft);

	if(strlen(xlp->str) > xlp->nchars){
		cp = MALLOC("cstr", xlp->nchars + 1);
		strncpy(cp, xlp->str, xlp->nchars + 1);
		cp[xlp->nchars] = '\0';
		fprintf(fp, "{ \"%s\" bg=%d }\n", cp, xlp->bg);
		FREE("cstr", cp);
	}
	else
		fprintf(fp, "{ \"%s\" bg=%d }\n", xlp->str, xlp->bg);
}

static void 
xstr_free(struct xobject *xp){ 
	if(((struct xstr *)xp)->needfree)
		FREE("xstr/str", (void *)((struct xstr *)xp)->str);
	FREE(xobj_typestr(xp), xp); 
}

static struct xobject *
xstring_new_i(const char *str, int nchars, struct xattr *at, int bg, int alloc)
{
	int w = 0, h = -1;
	struct xstr *ret = MALLOC("xstr", sizeof(struct xstr));
	char *newstr;

	h = xattr_get_font_height(at);
	w = xattr_get_str_width(at, str, nchars);

	xo_set((struct xobject *)ret, XOBJTYPE_STRING, XOBJFLAG_ENABLE,
		    w, h, NULL, xstr_free, xstr_dump);

	if(alloc){
		newstr = MALLOC("xstr/str", nchars + 1);
		memcpy(newstr, str, nchars + 1);
		newstr[nchars] = '\0';
		ret->str = newstr;
		ret->needfree = 1;
	}
	else  {
		ret->str = str;
		ret->needfree = 0;
	}
	xpmt_set_attr(ret, at);
	ret->bg = bg;
	ret->nchars = nchars;
	return (struct xobject *)ret;
}

struct xobject *xstring_new(const char *str, int nchars, 
			    struct xattr *at, int bg)
{
	return xstring_new_i(str, nchars, at, bg, 1);
}
struct xobject *xstring_na_new(const char *str, int nchars, 
			      struct xattr *at, int bg)
{
	return xstring_new_i(str, nchars, at, bg, 0);
}

/*----- wstring ----*/
static void xwstr_draw(struct xobject *xp, xwindow *xwp, int x, int y)
{
	struct xwstr *xrp = (struct xwstr *)xp;
	if(xrp->bg)
		xw_draw_wstring_bg(xwp, xpmt_attr(xp), x, y, 
			xrp->str, xrp->nchars);
	else
		xw_draw_wstring(xwp, xpmt_attr(xp), x, y, 
			xrp->str, xrp->nchars);
}

static void xwstr_dump(struct xobject *xp, FILE *fp, int oft) {
	char *cp;
	struct xwstr *xlp = (struct xwstr *)xp;
	xo_dump(xp, fp, oft);

	if(strlen(xlp->str) > xlp->nchars * 2){
		cp = MALLOC("cstr", (xlp->nchars * 2) + 1);
		strncpy(cp, xlp->str, (xlp->nchars * 2) + 1);
		cp[xlp->nchars*2] = '\0';
		fprintf(fp, "{ \"%s\" bg=%d }\n", cp, xlp->bg);
		FREE("cstr", cp);
	}
	else
		fprintf(fp, "{ \"%s\" bg=%d }\n", xlp->str, xlp->bg);
}

static void 
xwstr_free(struct xobject *xp){ 
	if(((struct xstr *)xp)->needfree)
		FREE("xwstr/str", (void *)((struct xstr *)xp)->str);
	FREE(xobj_typestr(xp), xp); 
}

struct xobject *
xwstring_new_i(const char *wstr, int nchars, struct xattr *at, int bg, int alloc)
{
	int w = 0, h = -1;
	struct xwstr *ret = MALLOC("xwstr", sizeof(struct xwstr));
	char *newstr;

	h = xattr_get_font_height(at);
	/* w = xattr_get_wstr_width(at, wstr, nchars); */
	w = xattr_get_font_width(at) * nchars;

	xo_set((struct xobject *)ret, XOBJTYPE_WSTRING, XOBJFLAG_ENABLE,
		     w, h, NULL, xwstr_free, xwstr_dump);

	if(alloc){
		newstr = MALLOC("xwstr/str", (nchars * 2)+ 1);
		memcpy(newstr, wstr, nchars * 2 + 1);
		newstr[nchars * 2] = '\0';
		ret->str = newstr;
		ret->needfree = 1;
	}
	else {
		ret->str = wstr;
		ret->needfree = 0;
	}
		
	xpmt_set_attr(ret, at);
	ret->bg = bg;
	ret->nchars = nchars;
	return (struct xobject *)ret;
}

struct xobject *xwstring_new(const char *wstr, int nchars, 
			    struct xattr *at, int bg)
{
	return xwstring_new_i(wstr, nchars, at, bg, 1);
}
struct xobject *xwstring_na_new(const char *wstr, int nchars, 
			      struct xattr *at, int bg)
{
	return xwstring_new_i(wstr, nchars, at, bg, 0);
}

/*--- string editor ---*/
static void 
xeditstr_draw(struct xobject *xp, xwindow *xwp, int x, int y)
{
	struct xeditstr *xrp = (struct xeditstr *)xp;
	int    offset = 0;
	int    strwid = xattr_get_str_width(xpmt_attr(xp), 
					 xrp->str, strlen(xrp->str));

	if(xrp->layout & XCRULE_LEFT)
		offset = 1;	/* 1pixel margin */
	else if(xrp->layout & XCRULE_RIGHT)
		offset = xobj_width(xp)- strwid - 1;	/* 1pixel margin */
	else 
		offset = (xobj_width(xp) - strwid) / 2;

	/* clear first */
	/* xw_clear_region(xwp, x, y, xobj_width(xp), xobj_height(xp));  */

	/* add offset */
	x += offset;

	/* draw string before position mark */
	if(xrp->pos > 0){
		xw_draw_string(xwp, xpmt_attr(xp), x, y, xrp->str, xrp->pos);
		x += xattr_get_str_width(xpmt_attr(xp), xrp->str, xrp->pos);
	}

	/* draw cursor position */
	xw_draw_string_bg(xwp, xrp->xatrev, x, y, xrp->str + xrp->pos, 1);
	x += xattr_get_font_width(xpmt_attr(xp));

	/* draw after cursor */
	if(xrp->pos < strlen(xrp->str) - 1)
		xw_draw_string(xwp, xpmt_attr(xp), x, y, xrp->str + xrp->pos + 1, 
				strlen(xrp->str) - xrp->pos - 1);
}

static void 
xeditstr_dump(struct xobject *xp, FILE *fp, int oft)
{  
	struct xeditstr *xlp = (struct xeditstr *)xp;
	xo_dump(xp, fp, oft);
	fprintf(fp, "{ \"%s\" }\n", xlp->str);
}

static void 
xeditstr_free(struct xobject *xp)
{ 
	FREE("xeditstr/str", (void *)((struct xeditstr *)xp)->str);
	FREE("xeditstr", xp); 
}

struct xobject *
xeditstr_new(unsigned int width, xattr *at1, xattr *at2, 
	     unsigned char hlayout_rule)
{
	int w, h;
	struct xeditstr *ret = MALLOC("xeditstr", sizeof(struct xeditstr));
	char *str;

	if(width <= 0)
		abort();	/* must be greater than 0 */

	w = xattr_get_font_width(at1);
	h = xattr_get_font_height(at1);
	xo_set((struct xobject *)ret, XOBJTYPE_EDITSTR, XOBJFLAG_ENABLE,
		    width, h, NULL, xeditstr_free, xeditstr_dump);

	ret->maxchar = width / w;
	str = MALLOC("xeditstr/str", ret->maxchar + 1);
	ret->str = str;
	ret->layout = hlayout_rule;	
	ret->pos = 0;
	ret->nchar = 1;	/* including space */

	str[0] = ' ';
	str[1] = '\0';

	xpmt_set_attr(ret, at1);
	ret->xatrev = at2;	/* reversed font */
	return (struct xobject *)ret;
}
