/*
 *			xpopdic.c
 * $Id: xpopdic.c,v 1.31 2004/01/23 02:45:27 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>

#include "debug.h"
#include "memory.h"
#include "wordhist.h"
#include "property.h"
#include "dictbook.h"
#include "booklist.h"
#include "dictlink.h"
#include "subr.h"

#include "xwrap/bitmap.h"
#include "xwrap/xwindow.h"
#include "stk/stkbase.h"
#include "strset.h"
#include "xobjcache.h"
#include "xpopdic_resource.h"
#include "xpopdic_component.h"
#include "xpopdic_control.h"

/* static variables */
static char 			curword[256];	/* current searched word */
static xrootwindow		*rwp;		/* root window */
static struct xselection 	*xsel;		/* X "PRIMARY" selection */
static struct xpopdic_control 	*xpctrl;	/* xpopdic controller */
static sigset_t 		newmask, oldmask;  /* signal set */

/* static functions */
static int  fetchword(xrootwindow *, struct xselection *, char *, int);
static void xpopdic_popup(void);
static void xpopdic_alarm(int sig);
static void alarminit(void);
static void alarmblock(void);
static void alarmunblock(void);


/* 
 * xpopdic_start(xrootwindow *xwp, booklist books):
 */
void 
xpopdic_start(xrootwindow *xrwp, booklist books)
{
	struct xobject *
	xppage_top_new(struct xpopdic_control *cp);
	xobject *top;
	/* get xpopdic controller */
	xpctrl = xpcontrol_initialize(xrwp, books);

	alarminit();	/* setup singal sets */
	rwp = xrwp;
	xsel = xselection_new(rwp, XSEL_PRIMARY);

	/* get current word to prevent popup with a cache word */
	fetchword(rwp, xsel, curword, 256);

	while(xpctrl->do_quit == 0){
		char w[256];	/* tmporary buffer */

		/* sleep for interval */
		sleep(xpctrl->xpr->intval);

		/* 
		 *  fetch a word in the primary selection 
		 * and compare it with the previous word.
		 */
		if(fetchword(rwp, xsel, w, 256) > 0){
			/* different word ? */
		   	if(strcmp(curword, w) != 0){
				/* new word */
				strcpy(curword, w);

				/* try to popup */
				xpopdic_popup();
			}
		}
		else {
			/* buffer reset */
			curword[0] = '_'; 
			curword[1] = '\0';
		}
	}

	xselection_free(xsel);
	xpcontrol_finalize(xrwp, xpctrl);
}


static int 
isword(const char *p)
{
	if(strlen(p) == 0)
		return 0;
	while(*p){
		if(!isalpha(*p))
			return 0;
		++p;
	}
	return 1;
}

static int
fetchword(xrootwindow *xrwp, struct xselection *xslp, char *buf, int  blen)
{
	int n;

	/* read X selection */
	if((n = xselection_get(xslp, buf, blen)) > 0){
		buf[n] = '\0';
		if(isword(buf)){
			if(DEBUG_IS_ENABLED(DEBUG_BUFFER))
				printf("(xselection): %s\n", buf);
			return n;
		}
		return 0;
	}

	{
		const char *cp;
		/* read cutbuffer */
		cp = xroot_cutbuffer(xrwp, &n);
		if(n > 0){
			if(isword(cp)){
				if(DEBUG_IS_ENABLED(DEBUG_BUFFER))
					printf("(xcutbuffer): %s\n", cp);
				strncpy(buf, cp, blen);
				return n;
			}
		}
	}
	return 0;
}

static void 
alarminit(void)
{
	struct sigaction act;
	sigemptyset(&newmask);
	sigaddset(&newmask, SIGALRM);

	/* call sigprocmask() to get current mask */
	alarmblock();
	alarmunblock();

	/* set a signal handler */
	act.sa_handler = xpopdic_alarm;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	if(sigaction(SIGALRM, &act, NULL) < 0){
		perror("sigaction");
		exit(1);
	}
}
static void 
alarmblock(void)
{
	/* block alarm */
	if(sigprocmask(SIG_BLOCK, &newmask, &oldmask) < 0){
		perror("sigprocmask");
		exit(1);
	}

}
static void 
alarmunblock(void)
{
	/* restore */
	if(sigprocmask(SIG_SETMASK,  &oldmask, NULL) < 0){
		perror("sigprocmask");
		exit(1);
	}
}

static void 
xpopdic_alarm(int sig)
{ 
	char w[256];
	if(fetchword(rwp, xsel, w, 256) > 0 &&
	   strcmp(curword, w) != 0){
		/* new word */
		strcpy(curword, w);

		/* search a new word */ 
		xpc_search(xpctrl, curword);
	}
	alarm(xpctrl->xpr->intval);
}

static void 
xpopdic_popup(void)
{
	/* try to search */
	xpc_search(xpctrl, curword);

	/* found! */
	if(xpctrl->dn_cur){
		/* start alarm */
		alarm(xpctrl->xpr->intval);	

		/* popup */
		xpc_popup(xpctrl);

		/* get input focus */
		xw_focus(xpctrl->win);

		/* stk loop */
		stk_start_loop(alarmblock, alarmunblock);

		/* return input focus */
		xw_focus_return(xpctrl->win);

		/* reset alarm */
		alarm(0);
	}
}
